"""Auto-tuning hyper-parameters."""

import numpy as np

MIN_SEARCH_STEP = 1e-04


class AutoTune:
  """AutoTune Class.

  This auto-tuning method is implemented based on this paper:
  Park, Tae Jin, et al. "Auto-tuning spectral clustering for speaker
  diarization using normalized maximum eigengap." IEEE Signal Processing Letter
  2019.
  """

  def __init__(self,
               p_percentile_min=0.60,
               p_percentile_max=0.95,
               init_search_step=0.01,
               search_level=1):
    """Initialization of the autotune arguments.

    Args:
      p_percentile_min: minimum value of p_percentile
      p_percentile_max: maximum value of p_percentile
      init_search_step: initial search step size for auto-tuning
      search_level: hierarchical search level for auto-tuning
    """
    self.p_percentile_min = p_percentile_min
    self.p_percentile_max = p_percentile_max
    self.search_step = init_search_step
    self.search_level = search_level

  def get_percentile_range(self):
    """Get the current percentile search range."""
    num_steps = int(
        np.ceil(
            (self.p_percentile_max - self.p_percentile_min) / self.search_step))
    return list(
        np.linspace(self.p_percentile_min, self.p_percentile_max, num_steps))

  def update_percentile_range(self, p_percentile_min, p_percentile_max,
                              search_step):
    """Update the percentile search range."""
    self.p_percentile_min = p_percentile_min
    self.p_percentile_max = p_percentile_max
    self.search_step = search_step
    return self.get_percentile_range()

  def tune(self, p_percentile_to_ratio):
    """Tune the hyper-parameter p_percentile.

    Use a proxy ratio of DER to tune the hyper-parameter p_percentile. It also
    performs some side work to do affinity refinement, eigen decomposition, and
    estimate the number of clusters.

    Args:
      p_percentile_to_ratio: a callable to compute the `ratio` given a
        `p_percentile` value

    Returns:
      eigenvectors: sorted eigenvectors. numpy array of shape
      (n_samples, n_samples)
      n_clusters: number of clusters as an integer
      best_p_percentile: p_percentile value that minimizes the ratio
    """
    p_percentile_range = self.get_percentile_range()
    searched = dict()
    for _ in range(self.search_level):
      min_ratio = np.inf
      for index, p_percentile in enumerate(p_percentile_range):
        if p_percentile in searched:
          continue
        # ratio is a proxy value of DER. We minimize this ratio
        # to find a good p_percentile
        ratio, eigenvectors_p, n_clusters_p = p_percentile_to_ratio(
            p_percentile)
        searched[p_percentile] = ratio
        if ratio < min_ratio:
          min_ratio = ratio
          eigenvectors = eigenvectors_p
          n_clusters = n_clusters_p
          best_p_percentile = p_percentile
          best_p_percentile_index = index
      # If the search range is not valid or search step is too small, we stop
      if not p_percentile_range or len(
          p_percentile_range) == 1 or self.search_step < MIN_SEARCH_STEP:
        break
      # Update the search range of p_percentile.
      # We search again from `start_index` position to `end_index` position
      # which is `local_search_dist` away from the found
      # `best_p_percentile_index` position. `search_step` is reduced to half of
      # the original size
      local_search_dist = max(2, len(p_percentile_range) // 8)
      start_index = max(0, best_p_percentile_index - local_search_dist)
      end_index = min(
          len(p_percentile_range) - 1,
          best_p_percentile_index + local_search_dist)
      p_percentile_min = p_percentile_range[start_index]
      p_percentile_max = p_percentile_range[end_index]
      self.search_step = self.search_step / 2
      p_percentile_range = self.update_percentile_range(p_percentile_min,
                                                        p_percentile_max,
                                                        self.search_step)
    return eigenvectors, n_clusters, best_p_percentile



"""Constraint information."""
import abc
import enum
import numpy as np

EPS = 1e-10


class ConstraintName(enum.Enum):
  """The names of constrained operations."""
  # The Affinity Integration method
  AffinityIntegration = enum.auto()

  # The Constraint Propagation method
  ConstraintPropagation = enum.auto()


class IntegrationType(enum.Enum):
  """The integration types for the Affinity Integration method."""
  Max = enum.auto()
  Average = enum.auto()


class ConstraintOptions:
  """Constraint options for constrained clustering methods."""

  def __init__(self,
               constraint_name,
               apply_before_refinement,
               integration_type=None,
               constraint_propagation_alpha=0.6):
    """Initialization of the constraint arguments.

    Args:
      constraint_name: a ConstraintName. Name of the constrained clustering
        method
      apply_before_refinement: if True, this operation is applied before the
        affinity refinement. It is suggested to set as True for the
        ConstraintPropagation method and False for the AffinityIntegration
        method
      integration_type: a IntegrationType. Integration type for the Affinity
        Integration method
      constraint_propagation_alpha: alpha value of the constraint propagation
        method
    """
    if not isinstance(constraint_name, ConstraintName):
      raise TypeError("constraint_name must be a ConstraintName")
    elif constraint_name == ConstraintName.AffinityIntegration:
      self.constraint_operator = AffinityIntegration(integration_type)
    elif constraint_name == ConstraintName.ConstraintPropagation:
      self.constraint_operator = ConstraintPropagation(
          constraint_propagation_alpha)
    else:
      raise ValueError(
          "Unsupported constrained name: {}".format(constraint_name))
    self.apply_before_refinement = apply_before_refinement


class ConstraintOperation(metaclass=abc.ABCMeta):
  """Constraint operation class."""

  def check_input(self, affinity, constraint_matrix):
    """Check the input to the adjust_affinity method.

    Args:
      affinity: the input affinity matrix.
      constraint_matrix: numpy array of shape (n_samples, n_samples). The
        constraint matrix with prior information

    Raises:
      TypeError: if affinity or constraint matrix has wrong type
      ValueError: if affinity or constraint matrix has wrong shape, etc.
    """
    if not isinstance(affinity, np.ndarray):
      raise TypeError("affinity must be a numpy array")
    if not isinstance(constraint_matrix, np.ndarray):
      raise TypeError("constraint matrix must be a numpy array")
    if len(affinity.shape) != 2:
      raise ValueError("affinity must be 2-dimensional")
    if affinity.shape[0] != affinity.shape[1]:
      raise ValueError("affinity must be a square matrix")
    if len(constraint_matrix.shape) != 2:
      raise ValueError("constraint matrix must be 2-dimensional")
    if constraint_matrix.shape[0] != constraint_matrix.shape[1]:
      raise ValueError("constraint matrix must be a square matrix")
    if affinity.shape != constraint_matrix.shape:
      raise ValueError(
          "affinity and constraint matrix must have the same shape")

  @abc.abstractmethod
  def adjust_affinity(self, affinity, constraint_matrix):
    """An abstract method to perform the constraint operation.

    Args:
      affinity: the affinity matrix, of size (n_samples, n_samples)
      constraint_matrix: numpy array of shape (n_samples, n_samples). The
        constraint matrix with prior information

    Returns:
      a matrix of the same size as affinity
    """
    pass


class AffinityIntegration(ConstraintOperation):
  """The Affinity Integration method.

  Basic operations to integrate the affinity matrix with given pairwise
  constraints in the constraint matrix. Current integration types include `Max`
  and `Average`.
  """

  def __init__(self, integration_type=IntegrationType.Max):
    self.integration_type = integration_type

  def adjust_affinity(self, affinity, constraint_matrix):
    """Adjust the affinity matrix with constraints."""
    self.check_input(affinity, constraint_matrix)
    if not isinstance(self.integration_type, IntegrationType):
      raise TypeError("integration_type must be a IntegrationType")
    elif self.integration_type == IntegrationType.Max:
      return np.maximum(affinity, constraint_matrix)
    elif self.integration_type == IntegrationType.Average:
      return 0.5 * (affinity + constraint_matrix)
    else:
      raise ValueError("Unsupported integration type: {}".format(
          self.integration_type))


class ConstraintPropagation(ConstraintOperation):
  """The Constraint Propagation method.

  The pairwise constraints are firstly propagated throughout the whole graph by
  two independent horizontal and vertical propagations. The final propagated
  constraint matrix is applied to adjust the affinity matrix.

  Reference:
  [1] Lu, Zhiwu, and IP, Horace HS. "Constrained spectral clustering via
  exhaustive and efficient constraint propagation." ECCV 2010
  [2] Lu, Zhiwu, and Peng, Yuxin. "Exhaustive and efficient constraint
  propagation: A graph-based learning approach and its applications." IJCV 2013
  """

  def __init__(self, alpha=0.6):
    self.alpha = alpha

  def adjust_affinity(self, affinity, constraint_matrix):
    """Adjust the affinity matrix with constraints."""
    self.check_input(affinity, constraint_matrix)
    adjusted_affinity = np.copy(affinity)
    degree = np.diag(np.sum(affinity, axis=1))
    degree_norm = np.diag(1 / (np.sqrt(np.diag(degree)) + EPS))
    # Compute affinity_norm as D^(-1/2)AD^(-1/2)
    affinity_norm = degree_norm.dot(affinity).dot(degree_norm)
    # The closed form of the final converged constraint matrix is:
    # (1-alpha)^2 * (I-alpha*affinity_norm)^(-1) * constraint_matrix *
    # (I-alpha*affinity_norm)^(-1). We save (I-alpha*affinity_norm)^(-1) as a
    # `temp_value` for readibility
    temp_value = np.linalg.inv(
        np.eye(affinity.shape[0]) - self.alpha * affinity_norm)
    final_constraint_matrix = (
        1 - self.alpha)**2 * temp_value.dot(constraint_matrix).dot(temp_value)
    # `is_positive` is a mask matrix where values of the final_constraint_matrix
    # are positive. The affinity matrix is adjusted by the final constraint
    # matrix using equation (4) in refernce paper [1]
    is_positive = final_constraint_matrix > 0
    affinity1 = 1 - (1 - final_constraint_matrix * is_positive) * (
        1 - affinity * is_positive)
    affinity2 = (1 + final_constraint_matrix * np.invert(is_positive)) * (
        affinity * np.invert(is_positive))
    adjusted_affinity = affinity1 + affinity2
    return adjusted_affinity


class ConstraintMatrix:
  """Constraint Matrix class."""

  def __init__(self, speaker_turn_scores, threshold=1):
    """Initialization of the constraint matrix arguments.

    Args:
      speaker_turn_scores: A list of speaker turn confidence scores. All score
        values are larger or equal to 0. If score is 0, there is no speaker
        turn. speaker_turn_scores[i+1] means the speaker turn confidence score
        between turn i+1 and turn i. The first score speaker_turn_scores[0] is
        not used.
      threshold: A threshold value for the speaker turn confidence score.
    """
    if any(score < 0 for score in speaker_turn_scores):
      raise ValueError("Speaker turn score must be larger or equal to 0.")
    self.speaker_turn_scores = speaker_turn_scores
    self.threshold = threshold

  def compute_diagonals(self):
    """Compute diagonal constraint matrix."""
    num_turns = len(self.speaker_turn_scores)
    constraint_matrix = np.zeros((num_turns, num_turns))
    for i in range(num_turns - 1):
      speaker_turn_score = self.speaker_turn_scores[i + 1]
      if speaker_turn_score != 0:
        if speaker_turn_score > self.threshold:
          constraint_matrix[i, i + 1] = -1
          constraint_matrix[i + 1, i] = -1
      else:
        constraint_matrix[i, i + 1] = 1
        constraint_matrix[i + 1, i] = 1
    return constraint_matrix



"""Laplacian matrix."""

import enum
import numpy as np

EPS = 1e-10


class LaplacianType(enum.Enum):
  """Different types of Laplacian matrix."""
  # The affinity matrix, not a Laplacian: W
  Affinity = enum.auto()

  # The unnormalied Laplacian: L = D - W
  Unnormalized = enum.auto()

  # The random walk view normalized Laplacian:  D^{-1} * L
  RandomWalk = enum.auto()

  # The graph cut view normalized Laplacian: D^{-1/2} * L * D^{-1/2}
  GraphCut = enum.auto()


def compute_laplacian(affinity, laplacian_type=LaplacianType.GraphCut, eps=EPS):
  """Compute the Laplacian matrix.

  Args:
    affinity: the affinity matrix of input data
    laplacian_type: a LaplacianType
    eps: a small value for numerial stability

  Returns:
    the Laplacian matrix

  Raises:
    TypeError: if laplacian_type is not a LaplacianType
    ValueError: if laplacian_type is not supported
  """
  degree = np.diag(np.sum(affinity, axis=1))
  laplacian = degree - affinity
  if not isinstance(laplacian_type, LaplacianType):
    raise TypeError("laplacian_type must be a LaplacianType")
  elif laplacian_type == LaplacianType.Affinity:
    return affinity
  elif laplacian_type == LaplacianType.Unnormalized:
    return laplacian
  elif laplacian_type == LaplacianType.RandomWalk:
    # Random walk normalized version
    degree_norm = np.diag(1 / (np.diag(degree) + eps))
    laplacian_norm = degree_norm.dot(laplacian)
    return laplacian_norm
  elif laplacian_type == LaplacianType.GraphCut:
    # Graph cut normalized version
    degree_norm = np.diag(1 / (np.sqrt(np.diag(degree)) + eps))
    laplacian_norm = degree_norm.dot(laplacian).dot(degree_norm)
    return laplacian_norm
  else:
    raise ValueError("Unsupported laplacian_type.")



"""Utility functions."""

import enum
import numpy as np

EPS = 1e-10


class EigenGapType(enum.Enum):
  """Different types of the eigengap computation."""
  # Eigengap is the ratio of two eigenvalues
  Ratio = enum.auto()

  # Eigengap is the subtraction of two eigenvalues, and it is normalized
  # by the maximum eigenvalue
  NormalizedDiff = enum.auto()


def compute_affinity_matrix(embeddings):
  """Compute the affinity matrix from data.

  Note that the range of affinity is [0, 1].

  Args:
    embeddings: numpy array of shape (n_samples, n_features)

  Returns:
    affinity: numpy array of shape (n_samples, n_samples)
  """
  # Normalize the data.
  l2_norms = np.linalg.norm(embeddings, axis=1)
  embeddings_normalized = embeddings / l2_norms[:, None]
  # Compute cosine similarities. Range is [-1,1].
  cosine_similarities = np.matmul(embeddings_normalized,
                                  np.transpose(embeddings_normalized))
  # Compute the affinity. Range is [0,1].
  # Note that this step is not mentioned in the paper!
  affinity = (cosine_similarities + 1.0) / 2.0

  return affinity


def compute_sorted_eigenvectors(input_matrix, descend=True):
  """Sort eigenvectors by the real part of eigenvalues.

  Args:
    input_matrix: the matrix to perform eigen analysis with shape (M, M)
    descend: sort eigenvalues in a descending order. Default is True

  Returns:
    w: sorted eigenvalues of shape (M,)
    v: sorted eigenvectors, where v[;, i] corresponds to ith largest
       eigenvalue
  """
  # Eigen decomposition.
  eigenvalues, eigenvectors = np.linalg.eig(input_matrix)
  eigenvalues = eigenvalues.real
  eigenvectors = eigenvectors.real
  if descend:
    # Sort from largest to smallest.
    index_array = np.argsort(-eigenvalues)
  else:
    # Sort from smallest to largest.
    index_array = np.argsort(eigenvalues)
  # Re-order.
  w = eigenvalues[index_array]
  v = eigenvectors[:, index_array]
  return w, v


def compute_number_of_clusters(eigenvalues,
                               max_clusters=None,
                               stop_eigenvalue=1e-2,
                               eigengap_type=EigenGapType.Ratio,
                               descend=True,
                               eps=EPS):
  """Compute number of clusters using EigenGap principle.

  Use maximum EigenGap principle to find the number of clusters.

  Args:
    eigenvalues: sorted eigenvalues of the affinity matrix
    max_clusters: max number of clusters allowed
    stop_eigenvalue: we do not look at eigen values smaller than this
    eigengap_type: the type of the eigengap computation
    descend: sort eigenvalues in a descending order. Default is True
    eps: a small value for numerial stability

  Returns:
    max_delta_index: number of clusters as an integer
    max_delta_norm: normalized maximum eigen gap
  """
  if not isinstance(eigengap_type, EigenGapType):
    raise TypeError("eigengap_type must be a EigenGapType")
  max_delta = 0
  max_delta_index = 0
  range_end = len(eigenvalues)
  if max_clusters and max_clusters + 1 < range_end:
    range_end = max_clusters + 1

  if not descend:
    # The first eigen value is always 0 in an ascending order
    for i in range(1, range_end - 1):
      if eigengap_type == EigenGapType.Ratio:
        delta = eigenvalues[i + 1] / (eigenvalues[i] + eps)
      elif eigengap_type == EigenGapType.NormalizedDiff:
        delta = (eigenvalues[i + 1] - eigenvalues[i]) / np.max(eigenvalues)
      else:
        raise ValueError("Unsupported eigengap_type")
      if delta > max_delta:
        max_delta = delta
        max_delta_index = i + 1  # Index i means i+1 clusters
  else:
    for i in range(1, range_end):
      if eigenvalues[i - 1] < stop_eigenvalue:
        break
      if eigengap_type == EigenGapType.Ratio:
        delta = eigenvalues[i - 1] / (eigenvalues[i] + eps)
      elif eigengap_type == EigenGapType.NormalizedDiff:
        delta = (eigenvalues[i - 1] - eigenvalues[i]) / np.max(eigenvalues)
      else:
        raise ValueError("Unsupported eigengap_type")
      if delta > max_delta:
        max_delta = delta
        max_delta_index = i

  return max_delta_index, max_delta


def enforce_ordered_labels(labels):
  """Transform the label sequence to an ordered form.

  This is the same type of label sequence used in the paper "Discriminative
  neural clustering for speaker diarisation". This makes the label sequence
  permutation invariant.

  Args:
    labels: an array of integers

  Returns:
    new_labels: an array of integers, where it starts with 0 and smaller
      labels always appear first
  """
  new_labels = labels.copy()
  max_label = -1
  label_map = {}
  for element in labels.tolist():
    if element not in label_map:
      max_label += 1
      label_map[element] = max_label
  for key in label_map:
    new_labels[labels == key] = label_map[key]
  return new_labels



"""Example configurations."""

from spectralcluster import autotune
from spectralcluster import constraint
from spectralcluster import laplacian
from spectralcluster import refinement
from spectralcluster import spectral_clusterer

AutoTune = autotune.AutoTune
ConstraintName = constraint.ConstraintName
ConstraintOptions = constraint.ConstraintOptions
RefinementName = refinement.RefinementName
RefinementOptions = refinement.RefinementOptions
ThresholdType = refinement.ThresholdType
SymmetrizeType = refinement.SymmetrizeType
LaplacianType = laplacian.LaplacianType
SpectralClusterer = spectral_clusterer.SpectralClusterer

# Configurations that are closest to the ICASSP2018 paper
# "Speaker Diarization with LSTM".
ICASSP2018_REFINEMENT_SEQUENCE = [
    RefinementName.CropDiagonal,
    RefinementName.GaussianBlur,
    RefinementName.RowWiseThreshold,
    RefinementName.Symmetrize,
    RefinementName.Diffuse,
    RefinementName.RowWiseNormalize,
]

icassp2018_refinement_options = RefinementOptions(
    gaussian_blur_sigma=1,
    p_percentile=0.95,
    thresholding_soft_multiplier=0.01,
    thresholding_type=ThresholdType.RowMax,
    refinement_sequence=ICASSP2018_REFINEMENT_SEQUENCE)

icassp2018_clusterer = SpectralClusterer(
    min_clusters=2,
    max_clusters=7,
    autotune=None,
    laplacian_type=None,
    refinement_options=icassp2018_refinement_options,
    custom_dist="cosine")

# Configurations of Turn-to-Diarize system using the
# Turn + Constraint Propagation + AutoTune method described in the paper
# "Turn-to-Diarize: Online Speaker Diarization Constrained by
# Transformer Transducer Speaker Turn Detection".
TURNTODIARIZE_REFINEMENT_SEQUENCE = [
    RefinementName.RowWiseThreshold, RefinementName.Symmetrize
]

turntodiarize_refinement_options = RefinementOptions(
    thresholding_soft_multiplier=0.01,
    thresholding_type=ThresholdType.Percentile,
    thresholding_with_binarization=True,
    thresholding_preserve_diagonal=True,
    symmetrize_type=SymmetrizeType.Average,
    refinement_sequence=TURNTODIARIZE_REFINEMENT_SEQUENCE)

turntodiarize_constraint_options = ConstraintOptions(
    constraint_name=ConstraintName.ConstraintPropagation,
    apply_before_refinement=True,
    constraint_propagation_alpha=0.4)

turntodiarize_auto_tune = AutoTune(
    p_percentile_min=0.40,
    p_percentile_max=0.95,
    init_search_step=0.05,
    search_level=1)

turntodiarize_clusterer = SpectralClusterer(
    min_clusters=2,
    max_clusters=7,
    refinement_options=turntodiarize_refinement_options,
    constraint_options=turntodiarize_constraint_options,
    autotune=turntodiarize_auto_tune,
    laplacian_type=LaplacianType.GraphCut,
    row_wise_renorm=True,
    custom_dist="cosine")



"""In some cases, we use a fallback clusterer instead of spectral.

Spectral clustering exploits the global structure of the data. But there are
cases where spectral clustering does not work as well as some other simpler
clustering methods, such as when the number of embeddings is too small.
"""

import enum
from sklearn.cluster import AgglomerativeClustering


class SingleClusterCondition(enum.Enum):
  """Which condition do we use for deciding single-vs-multi cluster(s)."""
  # If all affinities are larger than threshold, there is only a single cluster.
  AllAffinity = enum.auto()

  # If all neighboring affinities are larger than threshold, there is only
  # a single cluster.
  NeighborAffinity = enum.auto()

  # If the standard deviation of all affinities is smaller than threshold,
  # there is only a single cluster.
  AffinityStd = enum.auto()

  # Use fallback clusterer to make the decision. If fallback clusterer
  # finds multiple clusters, continue with spectral clusterer.
  # TODO: currently AgglomerativeClustering is hardcoded to 2 clusters, so this
  # is not useful. We will add other fallback clusterers in the future.
  FallbackClusterer = enum.auto()


class FallbackOptions:
  """Options for fallback options."""

  def __init__(self,
               spectral_min_embeddings=1,
               single_cluster_condition=SingleClusterCondition.AllAffinity,
               single_cluster_affinity_threshold=1.0):
    """Initialization of the fallback options.

    Args:
      spectral_min_embeddings: we only run spectral clusterer if we have at
        least these many embeddings; otherwise we run fallback clusterer
      single_cluster_condition: how do we decide single-vs-multi cluster(s)
      single_cluster_affinity_threshold: affinity threshold to decide
        whether there is only a single cluster
    """
    self.spectral_min_embeddings = spectral_min_embeddings
    self.single_cluster_condition = single_cluster_condition
    self.single_cluster_affinity_threshold = single_cluster_affinity_threshold


class FallbackClusterer:
  """Fallback clusterer. So far we simply use AgglomerativeClustering."""

  def __init__(self, options):
    """Initilization of the fallback clusterer.

    Args:
      options: an object of FallbackOptions
    """
    self.options = options
    self.clusterer = AgglomerativeClustering()

  def predict(self, embeddings):
    return self.clusterer.fit_predict(embeddings)
